/* $Id$

This file is part of gldbg
Copyright © 2010  Mikko Rasa, Mikkosoft Productions
Distributed under the GPL
*/

#include <msp/io/print.h>
#include "enums.h"
#include "gldbg.h"
#include "inspector.h"

using namespace std;
using namespace Msp;

Inspector::Inspector(GlDbg &dbg)
{
	CommandInterpreter &cmd_interp = dbg.get_command_interpreter();

	cmd_interp.register_command("state", this, &Inspector::cmd_state)
		.set_help("Inspects general GL state",
			"state vertex\n"
			"  Print current vertex attributes\n\n"
			"state bind\n"
			"  Show current bindings\n");

	cmd_interp.register_command("texture", this, &Inspector::cmd_texture)
		.set_help("Inspect texture state",
			"texture\n"
			"  Lists texture objects\n\n"
			"texture ID\n"
			"  Print information about a texture object\n");

	cmd_interp.register_command("buffer", this, &Inspector::cmd_buffer)
		.set_help("Inspect buffer object state",
			"buffer\n"
			"  Lists buffer objects\n\n"
			"buffer ID\n"
			"  Print information about a buffer object\n");
}

void Inspector::decode(const char *data, unsigned len)
{
	state.decode(data, len);
}

void Inspector::cmd_state(const string &args)
{
	const GlState &glstate = state;
	if(args=="vertex")
	{
		IO::print("Current vertex attributes:\n");
		const Vector4 &color = glstate.get_color();
		IO::print("  Color:     [%05.3f, %05.3f, %05.3f, %05.3f]\n", color.r, color.g, color.b, color.a);
		for(unsigned i=0; i<8; ++i)
		{
			const Vector4 &texcoord = glstate.get_texcoord(i);
			IO::print("  TexCoord%d: [%05.3f, %05.3f, %05.3f, %05.3f]\n", i, texcoord.s, texcoord.t, texcoord.p, texcoord.q);
		}
		const Vector3 &normal = glstate.get_normal();
		IO::print("  Normal:    [%05.3f, %05.3f, %05.3f]\n", normal.x, normal.y, normal.z);
	}
	else if(args=="bind")
	{
		IO::print("Current bindings:\n");
		for(unsigned i=0; i<8; ++i)
		{
			IO::print("  Texture unit %d:\n", i);
			const TexUnitState &unit = glstate.get_texture_unit(i);
			IO::print("    GL_TEXTURE_2D: %s\n", unit.describe_binding(GL_TEXTURE_2D));
			IO::print("    GL_TEXTURE_3D: %s\n", unit.describe_binding(GL_TEXTURE_3D));
		}
		IO::print("  Buffers:\n");
		const BufferState *buf = glstate.get_current_buffer(GL_ARRAY_BUFFER);
		IO::print("    GL_ARRAY_BUFFER:         %d\n", (buf ? buf->id : 0));
		buf = glstate.get_current_buffer(GL_ELEMENT_ARRAY_BUFFER);
		IO::print("    GL_ELEMENT_ARRAY_BUFFER: %d\n", (buf ? buf->id : 0));
	}
	else
		throw InvalidParameterValue("Invalid or missing argument");
}

void Inspector::cmd_texture(const string &args)
{
	if(args.empty())
	{
		const map<unsigned, TextureState> &textures = state.get_textures();
		IO::print("%d texture objects:\n", textures.size());
		for(map<unsigned, TextureState>::const_iterator i = textures.begin(); i!=textures.end(); ++i)
		{
			const TextureState &tex = i->second;
			IO::print("  %d: %s, %d images\n", i->first, tex.describe(), tex.images.size());
		}
	}
	else
	{
		unsigned id = lexical_cast<unsigned>(args);
		const TextureState &tex = state.get_texture(id);
		IO::print("Texture object %d\n", id);
		IO::print("  Target:          %s\n", describe_enum(tex.target, "TextureTarget"));
		IO::print("  Images:\n");
		for(unsigned i=0; i<tex.images.size(); ++i)
		{
			const TexImageState &img = tex.images[i];
			IO::print("    Level %2d:      %s\n", i, img.describe());
		}
		IO::print("  Min. filter:     %s\n", describe_enum(tex.min_filter, "TextureMinFilter"));
		IO::print("  Mag. filter:     %s\n", describe_enum(tex.mag_filter, "TextureMagFilter"));
		IO::print("  Wrap modes:      S=%s / T=%s / R=%s\n", describe_enum(tex.wrap_s, "TextureWrapMode"),
			describe_enum(tex.wrap_t, "TextureWrapMode"), describe_enum(tex.wrap_r, "TextureWrapMode"));
		IO::print("  Compare mode:    %s\n", describe_enum(tex.compare_mode, ""));
		IO::print("  Compare func:    %s\n", describe_enum(tex.compare_func, "DepthFunction"));
		IO::print("  Generate mipmap: %s\n", tex.generate_mipmap);
	}
}

void Inspector::cmd_buffer(const string &args)
{
	if(args.empty())
	{
		const GlState::BufferMap &buffers = state.get_buffers();
		IO::print("%d buffers:\n", buffers.size());
		for(GlState::BufferMap::const_iterator i=buffers.begin(); i!=buffers.end(); ++i)
			IO::print("  %d: %s\n", i->first, i->second.describe());
	}
	else
	{
		unsigned id = lexical_cast<unsigned>(args);
		const BufferState &buf = state.get_buffer(id);
		IO::print("Buffer %d:\n", id);
		IO::print("  Size:  %d bytes\n", buf.size);
		IO::print("  Usage: %s\n", describe_enum(buf.usage, ""));
		if(buf.content.arrays.size()==1 && buf.content.arrays.front().kind==GL_ELEMENT_ARRAY_BUFFER)
		{
			const BufferContent::Array &array = buf.content.arrays.front();
			IO::print("  Arrays:\n");
			IO::print("    0: Element indices, 1 %s\n", describe_enum(array.type, "DataType"));

			IO::print("  Data:\n");
			unsigned width = 1+buf.content.stride*2;
			string fmt = format(" %%%du", width);
			unsigned n_elems = buf.size/buf.content.stride;
			string line;
			const char *ptr = buf.data;
			for(unsigned i=0; i<n_elems; ++i)
			{
				if(line.empty())
					line = "   ";

				if(array.type==GL_UNSIGNED_BYTE)
					line += format(fmt, *(reinterpret_cast<const unsigned char *>(ptr)+i));
				else if(array.type==GL_UNSIGNED_SHORT)
					line += format(fmt, *(reinterpret_cast<const unsigned short *>(ptr)+i));
				else if(array.type==GL_UNSIGNED_INT)
					line += format(fmt, *(reinterpret_cast<const unsigned *>(ptr)+i));

				if(line.size()+1+width>79)
				{
					IO::print("%s\n", line);
					line.clear();
				}
			}

			if(!line.empty())
				IO::print("%s\n", line);
		}
		else if(buf.content.stride)
		{
			IO::print("  Stride: %d bytes\n", buf.content.stride);

			IO::print("  Arrays:\n");
			const vector<BufferContent::Array> &arrays = buf.content.arrays;
			for(vector<BufferContent::Array>::const_iterator i=arrays.begin(); i!=arrays.end(); ++i)
			{
				if(i->kind)
					IO::print("    %2d: %s, %d %s\n", i->offset,
						describe_enum(i->kind, ""), i->size, describe_enum(i->type, "DataType"));
				else
					IO::print("    %2d: Attrib %d, %d %s\n", i->offset,
						i->index, i->size, describe_enum(i->type, "DataType"));
			}

			IO::print("  Data:\n");
			string header;
			for(vector<BufferContent::Array>::const_iterator i=arrays.begin(); i!=arrays.end(); ++i)
			{
				if(!header.empty())
					header += " | ";

				string label;
				if(i->kind==GL_VERTEX_ARRAY)
					label = "Vertex";
				else if(i->kind==GL_NORMAL_ARRAY)
					label = "Normal";
				else if(i->kind==GL_COLOR_ARRAY)
					label = "Color";
				else if(i->kind==GL_TEXTURE_COORD_ARRAY)
				{
					if(i->size==1)
						label = "TexC";
					else
						label = "TexCoord";
				}
				else if(!i->kind)
				{
					if(i->size==1)
						label = format("A %d", i->index);
					else
						label = format("Attrib %d", i->index);
				}

				unsigned width = i->size;
				if(i->type==GL_FLOAT)
					width *= 5;
				else if(i->type==GL_UNSIGNED_BYTE)
					width *= 3;
				width += i->size-1;

				header.append((width-label.size())/2, ' ');
				header += label;
				header.append((width-label.size()+1)/2, ' ');
			}
			IO::print("     %s\n", header);

			unsigned n_verts = buf.size/buf.content.stride;
			for(unsigned i=0; i<n_verts; ++i)
			{
				const char *vertex = buf.data+i*buf.content.stride;

				string line;
				for(vector<BufferContent::Array>::const_iterator j=arrays.begin(); j!=arrays.end(); ++j)
				{
					if(!line.empty())
						line += " |";

					const char *base = vertex+j->offset;
					for(unsigned k=0; k<j->size; ++k)
					{
						if(j->type==GL_FLOAT)
							line += format(" %5.2f", *(reinterpret_cast<const float *>(base)+k));
						else if(j->type==GL_UNSIGNED_BYTE)
							line += format(" %3u", *(reinterpret_cast<const unsigned char *>(base)+k));
					}
				}

				IO::print("%3d:%s\n", i, line);
			}
		}
	}
}
